using System;
using System.Collections.Generic;
using UnityEngine;
using UnityEngine.Events;
using SimpleJSON;
using MVR.FileManagementSecure;

namespace Stopper {
	public class ControllerAdjust : MVRScript {
		
		string savePath = "Saves\\ControllerAdjust";
		string fileExt = "controlconfig";
		
		List<string> groups = new List<string>{"Arms and Shoulders", "Legs", "Torso", "Head", "Fingers", "Toes", "Genitalia"};
		
		Dictionary<string,string> jointNameToGroup = new Dictionary<string,string>{
			["head"] = "Head",
			["neck"] = "Head",
			["lowerJaw"] = "Head",
			
			
			["collar"] = "Arms and Shoulders",
			["shldr"] = "Arms and Shoulders",
			["forearm"] = "Arms and Shoulders",
			["hand"] = "Arms and Shoulders",
			
			["chest"] = "Torso",
			["abdomen"] = "Torso",
			["hip"] = "Torso",
			["pelvis"] = "Torso",
			
			["thigh"] = "Legs",
			["shin"] = "Legs",
			["foot"] = "Legs",
			
			
			["pinky"] = "Fingers",
			["ring"] = "Fingers",
			["mid"] = "Fingers",
			["index"] = "Fingers",
			["thumb"] = "Fingers",
			
			["toe"] = "Toes",

			["gen"] = "Genitalia",
		};
		
		
		Dictionary<JointConfig, JointConfig> jointSymmetryMapping = new Dictionary<JointConfig, JointConfig>();
		Dictionary<string, List<JointConfig>> groupToJoints = new Dictionary<string, List<JointConfig>>();
		
		List<JSONStorableFloat> limits = new List<JSONStorableFloat>();
		
		List<JointConfig> jointConfigs = new List<JointConfig>();
		
		public JSONStorableBool symmetricMode;
		
		public class JointConfig {
			public ControllerAdjust script;
			public string name;
			public string symmetryName;
			public string group;
			public Rigidbody rigidbody;
			public ConfigurableJoint joint;
			public FreeControllerV3 controller;
			public JointConfig symmetricJoint;
			
			public JSONStorableFloat xOffset;
			public JSONStorableFloat yOffset;
			public JSONStorableFloat zOffset;
			
			public UIDynamicSlider xOffsetSlider;
			public UIDynamicSlider yOffsetSlider;
			public UIDynamicSlider zOffsetSlider;

			public JSONStorableFloat xRotation;
			public JSONStorableFloat yRotation;
			public JSONStorableFloat zRotation;

			public UIDynamicSlider xRotationSlider;
			public UIDynamicSlider yRotationSlider;
			public UIDynamicSlider zRotationSlider;
			
			private bool uiLoaded = false;
			public bool visible = false;
			public bool ignoreSymmetry = false;
			
			
			public void Default(){
				ignoreSymmetry = true;
				xOffset.SetValToDefault();
				yOffset.SetValToDefault();
				zOffset.SetValToDefault();
				ignoreSymmetry = false;
			}
		
			public void Show(bool show){
				visible = show;
				if(!uiLoaded && show){
					xOffsetSlider = script.AddSlider(xOffset,
					(float angle) => {
						if(script.symmetricMode.val && !ignoreSymmetry && symmetricJoint != null && symmetricJoint.xOffset.val != -angle){
							symmetricJoint.xOffset.val = -angle;
						}
					}, true);

                    yOffsetSlider = script.AddSlider(yOffset,
					(float angle) => {
						if(script.symmetricMode.val && !ignoreSymmetry && symmetricJoint != null && symmetricJoint.yOffset.val != angle){
							symmetricJoint.yOffset.val = angle;
						}
					}, true);

                    zOffsetSlider = script.AddSlider(zOffset,
					(float angle) => {
						if(script.symmetricMode.val && !ignoreSymmetry && symmetricJoint != null && symmetricJoint.zOffset.val != angle){
							symmetricJoint.zOffset.val = angle;
						}
					}, true);

					xRotationSlider = script.AddSlider(xRotation,
					(float angle) => {
						if(script.symmetricMode.val && !ignoreSymmetry && symmetricJoint != null && symmetricJoint.xRotation.val != angle){
							symmetricJoint.xRotation.val = angle;
						}
					}, true);

					yRotationSlider = script.AddSlider(yRotation,
					(float angle) => {
						if(script.symmetricMode.val && !ignoreSymmetry && symmetricJoint != null && symmetricJoint.yRotation.val != angle){
							symmetricJoint.yRotation.val = -angle;
						}
					}, true);

					zRotationSlider = script.AddSlider(zRotation,
					(float angle) => {
						if(script.symmetricMode.val && !ignoreSymmetry && symmetricJoint != null && symmetricJoint.zRotation.val != angle){
							symmetricJoint.zRotation.val = -angle;
						}
					}, true);
				
					
					uiLoaded = true;
					return;
				}
				
				if(uiLoaded){
					xOffsetSlider.gameObject.SetActive(show);
					yOffsetSlider.gameObject.SetActive(show);
					zOffsetSlider.gameObject.SetActive(show);

					xRotationSlider.gameObject.SetActive(show);
					yRotationSlider.gameObject.SetActive(show);
					zRotationSlider.gameObject.SetActive(show);
				}
			}
		}
		
		public override void Init() {
			
			var geometry = containingAtom.GetStorableByID("PhysicsModel");
			
			if(geometry == null){
				SuperController.LogError("Could not find character geometry");
				return;
			}
			
			// Preset code was adapted from AcidBubbles' Collider Editor
			var loadPresetUI = CreateButton("Load Preset");
			loadPresetUI.button.onClick.AddListener(() =>
			{
				FileManagerSecure.CreateDirectory(savePath);
				var shortcuts = FileManagerSecure.GetShortCutsForDirectory(savePath);
				SuperController.singleton.GetMediaPathDialog(LoadPreset, fileExt, savePath, false, true, false, null, false, shortcuts);
			});

			var savePresetUI = CreateButton("Save Preset");
			savePresetUI.button.onClick.AddListener(() =>
			{
				FileManagerSecure.CreateDirectory(savePath);
				var fileBrowserUI = SuperController.singleton.fileBrowserUI;
				fileBrowserUI.SetTitle("Save Contortionlist Preset");
				fileBrowserUI.fileRemovePrefix = null;
				fileBrowserUI.hideExtension = false;
				fileBrowserUI.keepOpen = false;
				fileBrowserUI.fileFormat = fileExt;
				fileBrowserUI.defaultPath = savePath;
				fileBrowserUI.showDirs = true;
				fileBrowserUI.shortCuts = null;
				fileBrowserUI.browseVarFilesAsDirectories = false;
				fileBrowserUI.SetTextEntry(true);
				fileBrowserUI.Show(SavePreset);
				fileBrowserUI.fileEntryField.text = (int)(DateTime.UtcNow - new DateTime(1970, 1, 1)).TotalSeconds + "." + fileExt;
				fileBrowserUI.ActivateFileNameField();
			});
			
			var defaultAllButton = CreateButton("Default Shown", true);
			defaultAllButton.button.onClick.AddListener(() => {
				foreach(var jointConfig in jointConfigs){
					if(jointConfig.visible){
						jointConfig.Default();
					}
				}
			});
			
			
			symmetricMode = new JSONStorableBool("Symmetric Mode", true);
			var symmetricModeToggle = CreateToggle(symmetricMode, false);
			
			foreach(Rigidbody rb in geometry.gameObject.GetComponentsInChildren<Rigidbody>(true)){
				
				if(rb.name.ToLower().Contains("collider"))
					continue;
				
				bool configurable = false;
				
				string group = "";
				
				foreach(var entry in jointNameToGroup){
					string jointName = entry.Key;
					if(rb.name.ToLower().Contains(jointName)){
						configurable = true;
						group = entry.Value;
					}
				}
				
				if(!configurable) 
					continue;
				
                ConfigurableJoint joint = null;
				FreeControllerV3 controller = null;
                foreach(var j in rb.GetComponents<ConfigurableJoint>()){

					controller = j.connectedBody.GetComponent<FreeControllerV3>();
					if(controller != null){
						joint = j;
						break;
					}
                }

                if(joint == null)
					continue;
				
				JointConfig jointConfig = new JointConfig();
				jointConfig.script = this;
				
				jointConfig.group = group;
				
				jointConfig.rigidbody = rb;
				jointConfig.name = rb.name;
				jointConfig.symmetryName = joint.name.ToLower().Replace("r", "x").Replace("l", "x");
				
				jointConfig.joint = joint;
				jointConfig.controller = controller;
				
				jointConfig.xOffset = CreateFloat($"{rb.name} X Offset", 0f, -1f, 1f);
                jointConfig.yOffset = CreateFloat($"{rb.name} Y Offset", 0f, -1f, 1f);
                jointConfig.zOffset = CreateFloat($"{rb.name} Z Offset", 0f, -1f, 1f);

				jointConfig.xOffset.setCallbackFunction = (float val) => {
					AdjustAnchor(jointConfig);
				};

				jointConfig.yOffset.setCallbackFunction = (float val) => {
					AdjustAnchor(jointConfig);
				};

				jointConfig.zOffset.setCallbackFunction = (float val) => {
					AdjustAnchor(jointConfig);
				};

				jointConfig.xRotation = CreateFloat($"{rb.name} X Rotation", 0f, -180f, 180f);
				jointConfig.yRotation = CreateFloat($"{rb.name} Y Rotation", 0f, -180f, 180f);
				jointConfig.zRotation = CreateFloat($"{rb.name} Z Rotation", 0f, -180f, 180f);

				jointConfig.xRotation.setCallbackFunction = (float val) => {
					AdjustRotation(jointConfig);
				};

				jointConfig.yRotation.setCallbackFunction = (float val) => {
					AdjustRotation(jointConfig);
				};

				jointConfig.zRotation.setCallbackFunction = (float val) => {
					AdjustRotation(jointConfig);
				};
                
				
				if(!groupToJoints.ContainsKey(group)){
					groupToJoints[group] = new List<JointConfig>();
				}
				groupToJoints[group].Add(jointConfig);
				
				
				var symmetricJoint = jointConfigs.Find((j) => j.symmetryName == jointConfig.symmetryName);
				if(symmetricJoint != null){
					jointConfig.symmetricJoint = symmetricJoint; 
					symmetricJoint.symmetricJoint = jointConfig; 
				}
				jointConfigs.Add(jointConfig);
			}
			
			jointConfigs.Sort((a,b) => a.name.CompareTo(b.name));
			
			foreach(var group in groups){
				CreateGroupButton(group);
			}
		}

		private void AdjustAnchor(JointConfig jointConfig){
			var oldAnchor = jointConfig.joint.anchor;
			jointConfig.joint.anchor = new Vector3(jointConfig.xOffset.val, jointConfig.yOffset.val, jointConfig.zOffset.val);

			jointConfig.controller.transform.position += jointConfig.controller.transform.rotation * (jointConfig.joint.anchor - oldAnchor);
		}

		private void AdjustRotation(JointConfig jointConfig){
			var oldRotation = jointConfig.joint.targetRotation;
			jointConfig.joint.targetRotation = Quaternion.Euler(new Vector3(jointConfig.xRotation.val, jointConfig.yRotation.val, jointConfig.zRotation.val));
			jointConfig.controller.transform.rotation *= Quaternion.Inverse(oldRotation) * jointConfig.joint.targetRotation;
		}
		
		private void DefaultAll(){
			foreach(var jointConfig in jointConfigs){
				jointConfig.Default();
			}
		}
		
		public void CreateGroupButton(string group){
			UIDynamicButton groupButton = CreateButton(group);
			groupButton.button.onClick.AddListener(() => {
				foreach(var jointConfig in jointConfigs){
					bool show = jointConfig.group == groupButton.label;
										
					jointConfig.Show(show);
				}
			});
		}
		
		public JSONStorableFloat CreateFloat(string name, float defaultVal, float min, float max) {
			JSONStorableFloat storableFloat = new JSONStorableFloat(name, defaultVal, min, max, true, true);
			limits.Add(storableFloat);
			storableFloat.storeType = JSONStorableParam.StoreType.Full;
			RegisterFloat(storableFloat);
			return storableFloat;
		}
		
		public UIDynamicSlider AddSlider(JSONStorableFloat storableFloat, UnityAction<float> onChange = null, bool rightSide = false)
		{
			UIDynamicSlider uiSlider = this.CreateSlider(storableFloat, rightSide);
			uiSlider.rangeAdjustEnabled = true;
			
			if(onChange != null){
				uiSlider.slider.onValueChanged.AddListener(onChange);
			}
			
			return uiSlider;
		}
		
		private void Start(){
			
		}
	
			
		private void OnDisable()
		{
			DefaultAll();
		}
		
		private void OnDestroy(){
		}
		
		private void LoadPreset(string path) {
			if (string.IsNullOrEmpty(path))
				return;
			
			var presetJson = (JSONClass)LoadJSON(path);
			
			RestoreFromJSON(presetJson, true, true, null, false);
			
			
		}
		
		private void SavePreset(string path) {
			SuperController.singleton.fileBrowserUI.fileFormat = null;
			if (string.IsNullOrEmpty(path)) {
				return;
			}

			if (!path.ToLower().EndsWith($".{fileExt}")) {
				path += $".{fileExt}";
			}

			SaveJSON(GetJSON(true,true,true), path);
		}
		
		private string GetPathToObject(GameObject go){
			 if (go.transform.parent == null)
				 return "/" + go.name;
			 return GetPathToObject(go.transform.parent.gameObject) + "/" + go.name;
		}
	}
}
